/*
 * XLife Copyright 1989 Jon Bennett jb7m+@andrew.cmu.edu, jcrb@cs.cmu.edu
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* 
Enhancements to #I format added by Paul Callahan (callahan@cs.jhu.edu).
New format is backward compatible with old format.
*/
#define WIDGET_SHELL "wish"
#define FN_PICK_WIDGET "/usr/share/xlife/widgets/xlife.l.tcl"
#define PICKED_FN_FILE "/tmp/xlife.fn.tmp"

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <ctype.h>
#include <dirent.h>
#include <stdlib.h>
#include <errno.h>
#include "defs.h"

/* pattern name of tentative population */
static char tentpat[PATNAMESIZ];

/* global header cell for load script */ 
static LoadReq *loadscript;
 
/* matrix for rotations */
static int rotxx[] = {1, 0,-1, 0}; 
static int rotxy[] = {0,-1, 0, 1}; 
static int rotyx[] = {0, 1, 0,-1}; 
static int rotyy[] = {1, 0,-1, 0}; 

void fileinit(void) {
/* set up initial search path for file loads */
    char *cp, **dirp = dirs;

    if (cp = getenv("LIFEPATH")) {
	cp = strtok(cp, ":");
	do {
	    if (dirp - dirs > MAXDIRS)
		fatal("Too many pattern libraries\n");
	    *dirp++ = cp;
            qdirs++;
	} while
	    (cp = strtok((char *)NULL, ":"));
    }
    else {
	*dirp++ = "./";
	*dirp++ = LIFEDIR;
        qdirs = 2;
    }

    /* initialize load script to empty (circular list) */
    loadscript = (LoadReq *)malloc(sizeof(LoadReq));
    loadscript->next=loadscript;

}

static char *gethome() {
    return getpwuid(getuid())->pw_dir;
}

char *checktilda(char *stng) {
    static char full[1024];
    struct passwd *pw;
    int i;

    if (stng[0] == '~') {
	i = 1;
	while (stng[i] != '/' && stng[i] != '\0')
	    full[i - 1] = stng[i], i++;
	if (i == 1)
	    strcpy(full, gethome());
	else {
	    full[i - 1] = '\0';
	    if (pw = getpwnam(full))
                strcpy(full, pw->pw_dir);
	    else
		announce(SYSERR);
	}
	strcat(full, stng + i);
    }
    else
	strcpy(full, stng);
    return full;
}

char *addlifeext(char *buf) {
    int len = strlen(buf);
    if (strcmp(buf + len - LOADEXTLEN, LOADEXT))
	return LOADEXT;
    return "";
}

void savefile(void) {
    FILE *savefl;
    char outbuf[100];

    strcpy(inpbuf, "Save to: ");
    minbuflen = 9;
    strcat(inpbuf, "./");
    DoExpose(INPUTWIN);
    getxstring();
    strcpy(outbuf, checktilda(inpbuf + 9));
    strcat(outbuf, addlifeext(outbuf));
    if (savefl = fopen(outbuf,"w"))  {
        fprintf(savefl, "##XLife\n");
	saveall(savefl, saveformat);
	fclose(savefl);
    }
    else
	announce(SYSERR);
}

char *seppatdir(char *filename, char *filefield) {
    char *ffptr, *colptr;
    /* Separates path from rest of pattern's name.  */

    /* Make sure "/" in pattern block name isn't confusing by temporarily
       eliminating ":" suffix */
    if (colptr = strchr(filename, ':')) *colptr = '\0';

    ffptr = strrchr(filename, '/');
    if (ffptr) {
	strcpy(filefield, ffptr + 1);
	*(ffptr + 1)='\0';
    }
    else
    {
	strcpy(filefield, filename);
	*(filename)='\0';
    } 

    /* restore ":" suffix */
    if (colptr) *colptr = ':';

    return filename;
}

static int breakreq() {
    int breakval=0;

    if (XCheckMaskEvent(disp, KeyPressMask, &event)) {
	XLookupString(&event.xkey, keybuf, 16, &ks, &kstat);
	breakval = keybuf[0] == 'X';
	keybuf[0] = '\0';
    }
    return breakval;
}

void forgetpattern(pattern *pp) {
/* forget information about the last generation's state in a context */
    tile *ptr;

    for (ptr = pp->tiles; ptr; ptr = ptr->next)
        if (!ptr->dead)
            forget(&ptr->cells);
}

#define divis(x,y) ((x)/(y)*(y)==(x))

void do_loadreq(LoadReq *loadqueue, pattern *pp) {
    int quitload = 0;
    char thispat[PATNAMESIZ], badpat[PATNAMESIZ], *cp;

    /*
     * This code used to try to strip LIFEDIR off the front of patname
     * before copying it to tentpat.  LIFEDIR doesn't exist anymore as such,
     * because we handle multiple automata types now.  Now we just copy the
     * basename of patname...I hope this is equivalent!
     */
    if ((cp = strrchr(loadqueue->patname, '/')) == (char *)NULL)
        cp = loadqueue->patname;
    else
        ++cp;
    strcpy(tentpat, cp);

    /* clear tentative hash table */
    initcells(pp);

    badpat[0] = '\0';
    while (loadqueue) {
        while (loadqueue->loadtime > pp->generations) {
            if (divis(loadqueue->loadtime - pp->generations, 100)) {
                sprintf(inpbuf,
                        "Generating: %ld steps left (type X to halt)",
                        loadqueue->loadtime - pp->generations);
                DoExpose(INPUTWIN);
            }
            generate(pp);
            if (quitload = breakreq())
                break;
        }
        forgetpattern(pp);      /* keeps incremental display from losing */
        strcpy(thispat, loadqueue->patname + loadqueue->relpath);
        sprintf(inpbuf,
                "Loading %s (type X to halt)", thispat);
        DoExpose(INPUTWIN);
        if (!do_loadfile(&loadqueue, pp)) {
            strcpy(badpat, thispat);
            free_loadscript();
            quitload = 1;
        }
        if (quitload = quitload || breakreq())
            break;
    }
    DoExpose(LIFEWIN);

    pp->generations = 0;
    /* echo name of last pattern that couldn't be read */
    if (strlen(badpat)) {
        int t;
        strcpy(inpbuf,"Can't load ");
        if ((t = strlen(cp = badpat) + 11 - inputwwidth/FONTWIDTH) > 0)
           cp += t + 1;  /* chops exceeded left part of filepath */
        strcat(inpbuf, cp);
        DoExpose(INPUTWIN);
    }
}


void add_loadreq(LoadReq **ploadqueue, unsigned long loadtime, char *patname, int relpath,
		coord_t hotx, coord_t hoty, int rxx, int rxy, int ryx, int ryy) {
    LoadReq *newreq;

    /* Find first entry where request can go while maintaining time order.
       A heap would have better theoretical time complexity, of course, but
       in practice, the list probably won't get too long.

       After loop, we have a pointer to the pointer that must be modified
       (Could be pointer to list itself). */

    while (*ploadqueue != NULL && (*ploadqueue)->loadtime < loadtime)
	ploadqueue = &((*ploadqueue)->next);

    /* Create new request block and load it up */
    newreq = (LoadReq *)malloc(sizeof(LoadReq));
    newreq->loadtime = loadtime;
    /* Silently truncates long file names--should probably tell user */
    strncpy(newreq->patname, patname, PATNAMESIZ);
    newreq->relpath = relpath;
    newreq->hotx = hotx;
    newreq->hoty = hoty;
    newreq->rxx = rxx;
    newreq->rxy = rxy;
    newreq->ryx = ryx;
    newreq->ryy = ryy;

    /* Insert new request in queue */
    newreq->next = *ploadqueue;
    *ploadqueue = newreq;
}

void parse_patname(char *patname, char *patfield) {
/* Breaks "<filename>:<pattern>" into two strings. */
    char *pfptr;

    pfptr = strchr(patname, ':');
    if (pfptr) {
	*pfptr = '\0';
	strcpy(patfield, pfptr + 1);
    }
    else patfield[0] = '\0';
}

#define F_L             0
#define F_LIF           1
#define F_MCL           2
#define F_RLE           3
#define F_CELLS         4
#define F_ERR           100

int detect_format(char *s) {
   char *p = strrchr(s, '.');
   if (p == 0 || p[1] == 'l' && p[2] == 0)
      return F_L;
   else if (!strcasecmp(p + 1, "lif"))
      return F_LIF;
   else if (!strcasecmp(p + 1, "mcl"))
      return F_MCL;
   else if (!strcasecmp(p + 1, "rle"))
      return F_RLE;
   else if (!strcasecmp(p + 1, "cells"))
      return F_CELLS;
   return F_L;
}

#define M_ABSOLUTE	0
#define M_RELATIVE	1
#define M_PICTURE	2
#define M_RLE		3
#define M_DELTA		4

void make_comment_line(char *cp) {
   while (isspace(*cp)) cp++;
   while (isspace(cp[strlen(cp) - 1]))
      cp[strlen(cp) - 1] = 0;
   if (strlen(cp) > 1) {
       strcpy(comments[numcomments], cp);
       if (numcomments < MAXCOMMENTS - 1)
            numcomments++;
   }
}

int rle_decode_line (pattern *context, char *cp, coord_t hotx, coord_t hoty,
            int xoff, int yoff, int rxx, int rxy, int ryx, int ryy, int *x, int *y) {
  register int sum, i;

  while (*cp) {
     sum = 0;
     while (*cp <= '9' && *cp >= '0')
        sum = sum*10 + *cp++ - '0';
     if (!sum) sum++;
     switch (*cp) {
          case '$':
             *y += sum;
             *x = 0;
             break;
          case '.':
          case 'b':
             *x += sum;
             break;
          case 'o':
             for (i = 0; i < sum; i++) {
                chgcell(context,
                   hotx + tx(xoff + *x, yoff + *y, rxx, rxy),
                   hoty + ty(xoff + *x, yoff + *y, ryx, ryy), 1);
                ++*x;
             }
          case '\n':
          case '\t':
          case ' ':
          case '\r':
             break;
          default:
             if (*cp > 'O' || *cp < 'A') return 1;
             for (i = 0; i < sum; i++) {
                chgcell(context,
                   hotx + tx(xoff + *x, yoff + *y, rxx, rxy),
                   hoty + ty(xoff + *x, yoff + *y, ryx, ryy), *cp - '@');
                ++*x;
             }
          }
     cp++;
  }
  return 0;
}

int do_loadfile(LoadReq **loadqueue, pattern *context) {
    char patname[PATNAMESIZ], patfield[PATNAMESIZ], tmpstring[PATNAMESIZ];
    coord_t hotx, hoty;
    int relpath, rxx, rxy, ryx, ryy, state;
    FILE *loadfl;
    char patfile[BUFSIZ];
    LoadReq *tmpqptr;
    char fullpath[PATNAMESIZ], **dirp;
    int loadmode = M_ABSOLUTE, linect = 0, x, y, loadformat;

    fname[0] = 0;
    /* Get load request */
    strcpy(patname, (*loadqueue)->patname);
    relpath = (*loadqueue)->relpath;
    hotx = (*loadqueue)->hotx;
    hoty = (*loadqueue)->hoty;
    rxx = (*loadqueue)->rxx;
    rxy = (*loadqueue)->rxy;
    ryx = (*loadqueue)->ryx;
    ryy = (*loadqueue)->ryy;

    /* Delete request from queue */
    tmpqptr = (*loadqueue)->next; 
    free(*loadqueue);
    *loadqueue = tmpqptr;

    /* separate filename from pattern name */
    parse_patname(patname, patfield);

    if ((loadformat = detect_format(patname)) == F_L)
         /* add load extension to filename if needed */
         (void) strcat(patname, addlifeext(patname));

    /* open file */
    strcpy(patfile, patname);

    if (patfile[0] == '/')
	(void) strcpy(fullpath, patfile);
    else
    {
	for (dirp = dirs; *dirp; dirp++) {
	    DIR	*dp;
	    struct dirent *entry;

	    if (dp = opendir(*dirp)) {
		 while (entry = readdir(dp)) {
		     (void) sprintf(fullpath, "%s%s/%s",
				    *dirp, entry->d_name, patfile);
		     if (access(fullpath, R_OK) == 0) {
			 closedir(dp);
			 goto foundit;
		     }
		 }
		 closedir(dp);
	    }
	    else
	    {
		(void) sprintf(fullpath, "%s/%s", *dirp, patfile);
		if (access(fullpath, R_OK) == 0)
		    goto foundit;
	    }
	}
	return 0;
    }
foundit:
    if (loadfl = fopen(fullpath, "r")) {
      char    buf[BUFSIZ], patid[PATNAMESIZ], *cp;
      int     xoff = 0, yoff = 0, lx, ly;
      int     endpattern = 0, found = 0;

      outcome[0] = numcomments = 0;
      fseek(loadfl, 0, SEEK_SET);
      if (loadformat == F_L) {
        /* If we are searching for a specific pattern in the file,
           then we skip lines till we find it.  This is a pretty tacky way
           to handle multiple pattern definitions in the same file,
           but files with #I format probably won't get big enough for
           anyone to notice.  Really, we should implement a dictionary to
           save the location of a pattern definition in a file the first time
           we see it.
           */
        if (patfield[0] != '\0') {
            while (!found && fgets(buf, BUFSIZ, loadfl)) {
                if (buf[0] == '#' && buf[1] == 'B') {
                    for (cp = buf + 2; isspace(*cp) ||  *cp == ':'; cp++)
                            continue;
                    (void) sscanf(cp, "%s", patid);
                    found = !strcmp(patfield, patid);
                }
            }
            if (!found) {
                fclose(loadfl);
                return 0;
            }
        }
        while (fgets(buf, BUFSIZ, loadfl) && !endpattern) {
            if (buf[0] == '#') {
                char    incl[BUFSIZ];
                char    pardir[PATNAMESIZ], *errmsg;
                int     rotate, flip, relpathoffset = 0;
                unsigned long loadtime;

                incl[0] = '\0';
                switch(buf[1]) {
                case 'A':
                    loadmode = M_ABSOLUTE;
                    break;

                case 'B':
                    /* Anything between #B and #E is ignored, since it
                       is a pattern definition to be included later.
                       #B and #E cannot be nested, so we just skip till
                       we pass #E */
                    while (fgets(buf, BUFSIZ, loadfl)
                           && (buf[0] != '#' || buf[1] != 'E'));
                    break;

                case 'C':
                    make_comment_line(buf + 2);
                    break;

                case 'E':
                    /* if we hit a "#E" at this point, then it must be
                       the one that closes the "#B <patfield>" that we
                       are reading (assuming a syntactically correct file) */
                    endpattern = 1;
                    break;

                case 'I':
                    xoff = yoff = rotate = loadtime = 0;
                    flip = 1;
                    (void) sscanf(buf + 2, " %s %d %d %d %d %lu", incl,
                                  &xoff, &yoff, &rotate, &flip, &loadtime);
                    /* Silently ignore invalid flips */
                    if (flip != 1 && flip != -1) flip = 1;

                    /* if included pattern begins with ':' then assume
                       it's in active file */
                    if (incl[0] == ':') {
                        strcpy(tmpstring, patfile);
                        strcat(tmpstring, incl);
                        strcpy(incl, tmpstring);
                    }
                    else
                        /* if relative path given, assume directory of parent */
                        if (!strchr("/~", incl[0])) {
                            strcpy(pardir, patfile);
                            seppatdir(pardir, tmpstring);
                            relpathoffset = strlen(pardir);
                            strcat(pardir, incl);
                            strcpy(incl, pardir);
                        }
                    loadtime += context->generations;
                    if ((long)loadtime < 0) loadtime = 0;
                    add_loadreq(loadqueue, loadtime, incl, relpathoffset,
                                hotx + tx(xoff,yoff,rxx,rxy),
                                hoty + ty(xoff,yoff,ryx,ryy),
                                rxx*rotxx[mod(rotate,4)] + rxy*rotyx[mod(rotate,4)],
                                flip*(rxx*rotxy[mod(rotate,4)] + rxy*rotyy[mod(rotate,4)]),
                                ryx*rotxx[mod(rotate,4)] + ryy*rotyx[mod(rotate,4)],
                                flip*(ryx*rotxy[mod(rotate,4)] + ryy*rotyy[mod(rotate,4)]));
                    break;

                case 'K':
                    buf[strlen(buf) - 1] = '\0';
                    if (strlen(buf) > 3)
                         (void) strcpy(outcome, buf + 3);
                    break;

                case 'N':
                    (void) sscanf(buf + 2, " %s", fname);
                    break;

                case 'D':
                    loadmode = M_DELTA;
                    x = y = xoff = yoff = 0;
                    break;

                case 'P':
                    loadmode = M_PICTURE;
                    xoff = yoff = 0;
                    (void) sscanf(buf + 2, " %d %d", &xoff, &yoff);
                    linect = -1;
                    if (oldp) linect++;
                    break;

                case 'R':
                    loadmode = M_RELATIVE;
                    xoff = yoff = 0;
                    (void) sscanf(buf + 2, " %d %d", &xoff, &yoff);
                    break;

                case 'M':
                    x = y = xoff = yoff = 0;
                    loadmode = M_RLE;
                    break;

                case 'T':
                    for (cp = buf + 2; isspace(*cp); cp++)
                        continue;
                    cp[strlen(cp) - 1] = '\0';
                    if (file_rules(cp)) {
                       sprintf(inpbuf,
                              "Can't set entered rules (press a key)");
                       goto badrules;
                    }
                    break;

                case 'U':
#if STATEBITS > 1
                    for (cp = buf + 2; isspace(*cp); cp++)
                        continue;
                    cp[strlen(cp) - 1] = '\0';
                    if (strcmp(active_rules, cp))
                        if (errmsg = use_rules(cp)) {
                           sprintf(inpbuf,
                              "Can't load %s: %s (press a key)", cp, errmsg);
badrules:                  fclose(loadfl);
                           DoExpose(INPUTWIN);
                           wait_activity();
                           (void) file_rules("life");
                           return 0;
                        }
#endif
                    break;

                default:
                    break;
                }
            }
#if STATEBITS > 1
            else if (loadmode == M_ABSOLUTE && maxstates > 2
                     && sscanf(buf, "%d %d %d\n", &x, &y, &state) == 3
                     && state) {
                chgcell(context, xorigin + tx(x, y, rxx, rxy),
                        yorigin + ty(x, y, ryx, ryy), state);
                context->cellcount[state]++;
            }
#endif
            else if (loadmode == M_ABSOLUTE && sscanf(buf,"%d %d\n",&x,&y)==2) {
                chgcell(context,
                        xorigin + tx(x, y, rxx, rxy), yorigin + ty(x, y, ryx, ryy), 1);
                context->cellcount[1]++;
            }
#if STATEBITS > 1
            else if (loadmode == M_RELATIVE && maxstates > 2
                     && sscanf(buf, "%d %d %d\n", &x, &y, &state) == 3
                     && state)
            {
                chgcell(context,
                        hotx + tx(xoff + x, yoff + y, rxx, rxy),
                        hoty + ty(xoff + x, yoff + y, ryx, ryy),
                        state);
                context->cellcount[state]++;
            }
#endif
            else if (loadmode == M_RELATIVE && sscanf(buf,"%d %d\n", &x, &y) == 2) {
                chgcell(context,
                        hotx + tx(xoff + x, yoff + y, rxx, rxy),
                        hoty + ty(xoff + x, yoff + y, ryx, ryy),
                        1);
                context->cellcount[1]++;
            }
#if STATEBITS > 1
            else if (loadmode == M_DELTA && maxstates > 2
                     && sscanf(buf, "%d %d %d\n", &lx, &ly, &state) == 3
                     && state)
            {
                x += lx;
                y += ly;
                chgcell(context,
                        hotx + tx(xoff + x, yoff + y, rxx, rxy),
                        hoty + ty(xoff + x, yoff + y, ryx, ryy),
                        state);
                context->cellcount[state]++;
            }
#endif
            else if (loadmode == M_DELTA && sscanf(buf,"%d %d\n", &lx, &ly) == 2) {
                x += lx;
                y += ly;
                chgcell(context,
                        hotx + tx(xoff + x, yoff + y, rxx, rxy),
                        hoty + ty(xoff + x, yoff + y, ryx, ryy), 1);
                context->cellcount[1]++;
            }
            else if (loadmode == M_RLE)
               (void) rle_decode_line(context, buf, hotx, hoty, xoff, yoff, rxx, rxy, ryx, ryy, &x, &y);
            else                /* loadmode == M_PICTURE */
            {
                unsigned int bufoff = 0;

                for(;;) {
                    for (cp = buf; *cp; cp++)
                        if (*cp == '*') {
                            chgcell(context,
                                hotx + tx(xoff + bufoff + cp - buf, yoff + linect, rxx, rxy),
                                hoty + ty(xoff + bufoff + cp - buf, yoff + linect, ryx, ryy),
                                1);
                            context->cellcount[1]++;
                        }
#if STATEBITS > 1
                        else if (isdigit(*cp)) {
                            chgcell(context,
                                hotx + tx(xoff + bufoff + cp - buf, yoff + linect, rxx, rxy),
                                hoty + ty(xoff + bufoff + cp - buf, yoff + linect, ryx, ryy),
                                stoi(*cp));
                            context->cellcount[stoi(*cp)]++;
                        }
#endif                          /* STATEBITS */
                    if (*(cp - 1) == '\n')
                        break;
                    bufoff += BUFSIZ - 1;
                     if (!fgets(buf, BUFSIZ, loadfl))
                        break;
                }
            }
            linect++;
        }
      }
      else if (loadformat == F_LIF) {
         (void) file_rules("life");
         while (fgets(buf, BUFSIZ, loadfl)) {
            if (buf[0] == '#') {
                switch(buf[1]) {
                case 'D':
                    make_comment_line(buf + 2);
                    break;

                case 'N':
                    (void) sscanf(buf + 2, " %s", fname);
                    break;

                case 'P':
                    loadmode = M_PICTURE;
                    xoff = yoff = 0;
                    (void) sscanf(buf + 2, " %d %d", &xoff, &yoff);
                    linect = -1;
                    if (oldp) linect++;
                    break;

                case 'R':
                    for (cp = buf + 2; isspace(*cp); cp++)
                        continue;
                    cp[strlen(cp) - 1] = '\0';
                    if (file_rules(cp)) {
                       sprintf(inpbuf,
                              "Can't set rules %s (press a key)", cp);
                       goto badrules;
                    }
                    break;
                default:
                    break;
                }
            }
#if STATEBITS > 1
            else if (loadmode == M_ABSOLUTE && maxstates > 2
                     && sscanf(buf, "%d %d %d\n", &x, &y, &state) == 3
                     && state) {
                chgcell(context, xorigin + tx(x, y, rxx, rxy),
                        yorigin + ty(x, y, ryx, ryy), state);
                context->cellcount[state]++;
            }
#endif
            else if (loadmode == M_ABSOLUTE && sscanf(buf,"%d %d\n",&x,&y)==2) {
                chgcell(context,
                        xorigin + tx(x, y, rxx, rxy), yorigin + ty(x, y, ryx, ryy), 1);
                context->cellcount[1]++;
            }
            else                /* loadmode == M_PICTURE */
            {
                unsigned int bufoff = 0;

                while(1) {
                    for (cp = buf; *cp; cp++)
                        if (*cp == '*') {
                            chgcell(context,
                                hotx + tx(xoff + bufoff + cp - buf, yoff + linect, rxx, rxy),
                                hoty + ty(xoff + bufoff + cp - buf, yoff + linect, ryx, ryy),
                                1);
                            context->cellcount[1]++;
                        }
#if STATEBITS > 1
                        else if (isdigit(*cp)) {
                            chgcell(context,
                                hotx + tx(xoff + bufoff + cp - buf, yoff + linect, rxx, rxy),
                                hoty + ty(xoff + bufoff + cp - buf, yoff + linect, ryx, ryy),
                                stoi(*cp));
                            context->cellcount[stoi(*cp)]++;
                        }
#endif                          /* STATEBITS */
                    if (*(cp - 1) == '\n')
                        break;
                    bufoff += BUFSIZ - 1;
                    if (!fgets(buf, BUFSIZ, loadfl))
                        break;
                }
            }
            linect++;
        }
      }
      else if (loadformat == F_RLE) {
         x = y = xoff = yoff = 0;
         while (fgets(buf, BUFSIZ, loadfl)) {
            if (buf[0] == '#' || endpattern)
               make_comment_line(buf + 2 - endpattern);
            else if (sscanf(buf, " x = %d, y = %d, rule = %s ", &lx, &ly, patid) == 3) {
               if (cp = strrchr(patid, ':')) *cp = 0;
               strcpy(tmpstring, patid);
               if (cp = strchr(patid, '/'))
                  if (toupper(patid[0]) == 'S') {
                     strcpy(tmpstring, patid + 1);
                     tmpstring[cp - patid - 1] = '/';
                     strcpy(tmpstring + (cp - patid), cp + 2);
                  }
                  else if (toupper(patid[0]) == 'B') {
                     strcpy(tmpstring, cp + 2);
                     strcat(tmpstring, "/");
                     strncat(tmpstring, patid + 1, cp - patid - 1);
                  }
               if (file_rules(tmpstring) && use_rules(strcat(tmpstring, ".r"))) {
                  sprintf(inpbuf,
                              "Can't set rules %s (press a key)", tmpstring);
                  goto badrules;
               }
            }
            else if (rle_decode_line(context, buf, hotx, hoty, xoff, yoff, rxx, rxy, ryx, ryy, &x, &y))
                  endpattern = 2;
        }
      }
      else if (loadformat == F_CELLS) {
         (void) file_rules("life");
         linect = xoff = yoff = 0;
         while (fgets(buf, BUFSIZ, loadfl)) {
            if (buf[0] == '!')
                    make_comment_line(buf);
            else
            {
                unsigned int bufoff = 0;

                while(1) {
                    for (cp = buf; *cp; cp++)
                        if (*cp == 'O') {
                            chgcell(context,
                                hotx + tx(xoff + bufoff + cp - buf, yoff + linect, rxx, rxy),
                                hoty + ty(xoff + bufoff + cp - buf, yoff + linect, ryx, ryy),
                                1);
                            context->cellcount[1]++;
                        }
                    if (*(cp - 1) == '\n')
                        break;
                    bufoff += BUFSIZ - 1;
                    if (!fgets(buf, BUFSIZ, loadfl))
                        break;
                }
            }
            linect++;
        }
      }
      else if (loadformat == F_MCL) {
         int norules = 1;

         x = y = xoff = yoff = 0;
         while (fgets(buf, BUFSIZ, loadfl)) {
            if (!strncmp(buf, "#D", 2))
               make_comment_line(buf + 2);
            else if (norules && sscanf(buf, "#GOLLY %s ", tmpstring) == 1) {
               if (!file_rules(tmpstring))
                 norules = 0;
            }
            else if (norules && sscanf(buf, "#RULE %s ", patid) == 1) {
               strcpy(tmpstring, patid);
               if (file_rules(tmpstring) && use_rules(strcat(tmpstring, ".r"))) {
                  sprintf(inpbuf,
                              "Can't set rules %s (press a key)", tmpstring);
                  goto badrules;
               }
               else
                  norules = 1;
            }
            else if (!strncmp(buf, "#L ", 3))
               (void) rle_decode_line(context, buf + 3, hotx, hoty, xoff, yoff, rxx, rxy, ryx, ryy, &x, &y);
            else
               make_comment_line(buf);
        }
      }
      fclose(loadfl);
      return 1;
    }
    else
        return 0;
}

static void get_rot_flip(int rxx, int rxy, int ryx, int ryy, int *rot, int *flip) {
    /* compute rotation and flip from given transformation */
    /* No fancy mathematics; just search table of rotations */ 
    for (*rot = 0; *rot < 4; (*rot)++) {
	if (rxx == rotxx[*rot] && ryx == rotyx[*rot]) {
	    if (rxy == -rotxy[*rot] && ryy == -rotyy[*rot]) *flip = -1;
	    else *flip = 1;
	    break;
	}
    }
}

void free_loadscript(void) {
/* reinitialize load script to one self-connected cell */
    LoadReq *nptr,*ptr;

    for (ptr = loadscript->next; ptr != loadscript;) {
	nptr = ptr->next;
	free(ptr);
	ptr = nptr;
    }
    loadscript->next = loadscript; 
}

void saveloadscript(void) {
    FILE *savefl;
    char outbuf[100];
    static char savemsg[] = "Save load script to:";
    int rot = 0, flip = 0;
    unsigned long mingener;
    coord_t savex, savey;
    LoadReq *ptr;

    /* Confirm latest load before saving script */
    confirmload();
    DoExpose(LIFEWIN);

    if (loadscript == loadscript->next) {
        announce("There is no script to save!");
        return;
    }

    /* get name of file to write into into */
    strcpy(inpbuf, savemsg);
    minbuflen = strlen(savemsg);
    strcat(inpbuf, "./");
    DoExpose(INPUTWIN);

    getxstring();
    XClearWindow(disp, inputw);

    savex = XPOS(event.xmotion.x,xpos);
    savey = YPOS(event.xmotion.y,ypos);

    strcpy(outbuf, checktilda(inpbuf + strlen(savemsg)));
    strcat(outbuf, addlifeext(outbuf));
    if (savefl = fopen(outbuf,"w")) {
       fprintf(savefl, "##XLife\n");
	/* write load script */ 
	/* first, find minimum generation at which a pattern is
	   to be included so include times can be normalized */
        mingener = active.generations;
	for (ptr = loadscript->next; ptr != loadscript;) {
	    ptr = ptr->next;
	    if (mingener > ptr->loadtime) mingener = ptr->loadtime;
	}
	/* now output file */
	for (ptr = loadscript->next; ptr != loadscript;) {
	    ptr = ptr->next;
            get_rot_flip(ptr->rxx, ptr->rxy, ptr->ryx, ptr->ryy, &rot, &flip);
	    fprintf(savefl, "#I %s %ld %ld %d %d %ld\n",
		    ptr->patname, 
		    ptr->hotx - savex, 
		    ptr->hoty - savey, 
		    rot, flip, ptr->loadtime - mingener);
	}
	/* close file */
	fclose(savefl);
    }
    else
	announce(SYSERR);
}

static void add_include_entry(char *patname, long loadtime, coord_t hotx, coord_t hoty,
       int xx, int xy, int yx, int yy) {
    LoadReq *newreq;

    /* Create new include block and load it up */
    newreq = (LoadReq *)malloc(sizeof(LoadReq));
    strncpy(newreq->patname, patname, PATNAMESIZ);
    newreq->loadtime = loadtime;
    newreq->hotx = hotx;
    newreq->hoty = hoty;
    newreq->rxx = xx;
    newreq->rxy = xy;
    newreq->ryx = yx;
    newreq->ryy = yy;

    /* Insert new request at end of list */
    newreq->next = loadscript->next;
    loadscript->next = newreq;
    loadscript = newreq;
}

void copy_tentative(void) {
    tile *ptr;
    int dx, dy;
    long x, y;

    /* officially add loaded cells with given transformations */
    for (ptr = tentative.tiles; ptr; ptr = ptr->next) {
	int	state;

	if (!ptr->dead)
	    for (dx = 0; dx < BOXSIZE; dx++)
		for (dy = 0; dy < BOXSIZE; dy++)
		  if (state = getcell(&ptr->cells, dx, dy)) {
			x = ptr->x + dx - STARTX;
			y = ptr->y + dy - STARTY;
			chgcell(&active, tx(x, y, txx, txy) + loadx,
				ty(x, y, tyx, tyy) + loady, state);
		  }
    }
}

void confirmload(void) {
    if (tentative.tiles) {
        /* add completed load to load script */
	add_include_entry(tentpat, 
                          active.generations - tentative.generations,
			  loadx, loady, txx, txy, tyx, tyy);
        copy_tentative();

        /* clear the tentative pattern */
        clear_pattern(&tentative);

       /* undo boxing */
       boxpattern(0);
       adjhash(&active);
    }
}

void undoload(void) {
/* get rid of loaded cells */
    /* clear the tentative pattern */
    clear_pattern(&tentative);

    /* undo boxing */
    boxpattern(0);
}

void genload() {
    /* perform some generations on loaded cells */
    static char genmsg[] = "Number of generations to perform (default=1): ";
    int gens = 1;

    minbuflen = strlen(genmsg);
    announce(genmsg);
    getxstring();
    sscanf(inpbuf + strlen(genmsg), "%d", &gens);

    while (gens-- > 0) {
	if (divis(gens,100)) {
	    sprintf(inpbuf,
		    "Generating: %d steps left (type X to halt)", gens);
	    DoExpose(INPUTWIN);
	}
        if (tentative.tiles)
           generate(&tentative);
        else
	   generate(&active);
	if (breakreq())
	    break;
    }
}

void loadfile(void) {
    char outbuf[100], tmp[PATNAMESIZ];
    LoadReq *loadqueue = NULL;
    int status;
    FILE *fi;
    char s[256];

    confirmload();		/* confirm any previously loaded cells */ 
    if (stashed[0] != 0)
          strcpy(outbuf, stashed);
    else {
          strcpy(inpbuf, "Load from: ");
          minbuflen = 10;
          strcat(inpbuf, loadirbuf);
          if (fork() == 0) {
                remove(PICKED_FN_FILE);
                execlp(WIDGET_SHELL, WIDGET_SHELL, FN_PICK_WIDGET, 0);
                exit(1);
          }
          wait(&status);
          if (fi = fopen(PICKED_FN_FILE, "r")) {
                fgets(s, 256, fi);
                fclose(fi);
                if (!strcmp(s, "//")) return;
                strcpy(inpbuf + 11, s);
          }
          else
          {
                DoExpose(INPUTWIN);
                getxstring();
          }
          strcpy(outbuf, checktilda(inpbuf + 11));
          strcpy(stashed, outbuf);
    }
    DoExpose(LIFEWIN);
    /* redefine user load directory (but not pattern library) */
    strcpy(loadirbuf, outbuf);
    seppatdir(loadirbuf, tmp);

    /* don't need to add life extension now, since it will be added in
       do_loadfile  */

    iloadx = loadx = XPOS(event.xmotion.x, xpos);
    iloady = loady = YPOS(event.xmotion.y, ypos);
    txx = tyy = 1;
    txy = tyx = 0;
    add_loadreq(&loadqueue, 0, outbuf, 0, 
                STARTX, STARTY, 1, 0, 0, 1);
    dirs[qdirs] = loadirbuf; /* support for rules file */
    do_loadreq(loadqueue, &tentative);
    dirs[qdirs] = 0;

    if (cellcnt(&active) == 0) {
        unsigned cellwidth, cellheight;
        while (1) {
                cellwidth = shr(width - BORDERWIDTH*2, scale);
                cellheight = shr(height - INPUTH - BORDERWIDTH*3, scale);

                if (cellwidth >= tentative.xmax - tentative.xmin
                    && cellheight >= tentative.ymax - tentative.ymin || scale <= -10)
                    break;

                --scale;
                loadx = cellwidth/2 + xpos;
                loady = cellheight/2 + ypos;
        }
        setscale(scale);
        DoExpose(LIFEWIN);
    }
}

#if STATEBITS > 1
char* use_rules(char *rulefile) { /* load a given ruleset by name */
    char	*errmsg = 0, fullpath[PATNAMESIZ], **dirp;
    FILE	*fp;

    if (rulefile[0] == '/')
	(void) strcpy(fullpath, rulefile);
    else
    {
	for (dirp = dirs; *dirp; dirp++) {
	    DIR	*dp;
	    struct dirent *entry;

	    if (dp = opendir(*dirp)) {
		 while (entry = readdir(dp)) {
		     (void) sprintf(fullpath, "%s%s/%s",
				    *dirp, entry->d_name, rulefile);
		     if (access(fullpath, R_OK) == 0) {
			 closedir(dp);
			 goto foundit;
		     }
		 }
		 closedir(dp);
	    }
	    else
	    {
		(void) sprintf(fullpath, "%s/%s", *dirp, rulefile);
		if (access(fullpath, R_OK) == 0)
		    goto foundit;
	    }
	}
    }
 foundit:
    if ((fp = fopen(fullpath, "r")) == (FILE *)NULL)
	announce(errmsg = SYSERR);
    else if ((errmsg = readrules(fullpath)) == (char *)NULL) {
	(void) strcpy(active_rules, rulefile);
	showrules();
	DoExpose(INPUTWIN);	/* force the radio buttons on screen */
    }
    else
	announce(errmsg);
    return errmsg;
}

#define RULEPROMPT      "Load rules from: "
void loadrules(void) { /* load rules from a transition-definition file */
    char *errmsg, rules[80];

    if (state == STOP) {
	announce(RULEPROMPT);
	getxstring();
        strcpy(rules, inpbuf + strlen(RULEPROMPT));
        if (*rules != 0) {
             if (strcmp(rules + strlen(rules) - 2, ".r"))
                  strcat(rules, ".r");
             if (errmsg = use_rules(rules)) {
                  sprintf(inpbuf, "Can't load %s: %s (press a key)", rules, errmsg);
                  DoExpose(INPUTWIN);
                  wait_activity();
                  (void) file_rules("life");
             }
        }
        displaystats();
    }
}
#endif /* STATEBITS > 1 */
